import boto3
import io
import json
from typing import Optional, Any, Dict
from copy import copy
import os
import gzip
from enum import Enum
from my_aws_helpers.logging import select_powertools_logger


logger = select_powertools_logger("aws-helpers-s3")


class ContentType(str, Enum):
    plain_text = "text/plain"
    xml_content = "text/xml"
    json_content = "application/json"
    pdf_content = "application/pdf"


class ContentEncoding(str, Enum):
    gzip = "gzip"


class S3Location:
    bucket: str
    file_name: str
    location: str

    @classmethod
    def from_location(cls, location: str):
        bucket, file_name = location.split("/")[0], "/".join(location.split("/")[1:])
        return cls(bucket=bucket, file_name=file_name)

    def __init__(self, bucket: str, file_name: str) -> None:
        self.bucket = bucket
        self.file_name = file_name
        self.location = f"{self.bucket}/{self.file_name}"

    def serialise(self):
        return copy(vars(self))


class S3:
    client: boto3.client

    def __init__(
        self,
        client: Optional[boto3.client] = None,
        resource: Optional[boto3.resource] = None,
    ) -> None:
        self.client = client if client else self._get_client()
        self.resource = resource if resource else boto3.resource('s3')

    def _get_client(self) -> boto3.client:
        region_name = os.environ["AWS_DEFAULT_REGION"]
        s3_client = boto3.client("s3", region_name=region_name)
        endpoint_url = s3_client.meta.endpoint_url
        s3_client = boto3.client(
            "s3", region_name=region_name, endpoint_url=endpoint_url
        )
        return s3_client

    def _streaming_body_to_dict(self, payload):
        file_like_obj = io.BytesIO(payload.read())
        response = json.loads(file_like_obj.getvalue())
        return response

    def put_json_object(self, bucket_name: str, file_name: str, object: dict):
        return self.client.put_object(
            Body=json.dumps(object), Bucket=bucket_name, Key=file_name
        )

    def get_object(self, bucket_name: str, file_name: str):
        response = self.client.get_object(Bucket=bucket_name, Key=file_name)
        return self._streaming_body_to_dict(response["Body"])

    def get_presigned_url(
        self, bucket_name: str, file_name: str, expires_in: int = 3600
    ):
        return self.client.generate_presigned_url(
            "get_object",
            Params={
                "Bucket": bucket_name,
                "Key": file_name,
            },
            ExpiresIn=expires_in,
        )

    def get_s3_location_from_bucket_file(
        bucket_name: str, file_name: str
    ) -> S3Location:
        return S3Location(bucket=bucket_name, file_name=file_name)

    def get_bucket_file_from_s3_location(s3_location: S3Location) -> S3Location:
        return S3Location.from_location(location=s3_location)

    def save_document_content(
        self,
        file_contents: bytes,
        s3_location: S3Location,
        content_encoding: str = "",
        content_type: str = "application/pdf",
        compress: bool = True,
    ) -> Optional[S3Location]:
        """
        saves document content to bucket, in file_name
        Options for content_type:
            "application/pdf"
            "text/plain"
            "application/json"
            probably more
        Options for content_encoding:
            "": default encoding
            "gzip": compressed contents
        """
        try:
            if compress or s3_location.file_name.endswith(".gz"):
                file_contents = gzip.compress(file_contents)
                content_encoding = ContentEncoding.gzip.value
            obj = self.resource.Object(s3_location.bucket, s3_location.file_name)
            obj.put(
                Body=file_contents,
                ContentType=content_type,
                ContentEncoding=content_encoding,
            )
            return s3_location
        except Exception as e:
            logger.exception(e)
            return None        

    def read_binary_from_s3(self, s3_location: S3Location) -> Optional[bytes]:
        try:
            obj = self.resource.Object(s3_location.bucket, s3_location.file_name)
            d_bytes = io.BytesIO()
            obj.download_fileobj(d_bytes)
            d_bytes.seek(0)
            if obj.content_encoding == ContentEncoding.gzip.value:
                try:
                    with gzip.GzipFile(fileobj=d_bytes) as gz_file:
                        return gz_file.read()
                except gzip.BadGzipFile:
                    d_bytes.seek(0)
            return d_bytes.read()
        except Exception as e:
            logger.exception(f"Failed to read binary from s3 due to {e}")
            return None

    def save_text_to_s3(self, text: str, s3_location: S3Location):
        try:
            file_contents = bytes(text.encode("UTF-8"))
            return self.save_document_content(
                file_contents=file_contents,
                s3_location=s3_location,
                content_type=ContentType.plain_text.value,
                compress=True,
                content_encoding=ContentEncoding.gzip.value,
            )
        except Exception as e:
            logger.exception(f"Failed to save text to s3 due to {e}")
            return None

    def save_xml_to_s3(self, xml_text: bytes, s3_location: S3Location):
        """
        xml_text tends to come from:
            root = lxml.etree.ElementTree().get_root()
            xml_text = ET.tostring(root, encoding='utf-8')
        """
        try:
            return self.save_document_content(
                file_contents=xml_text,
                s3_location=s3_location,
                content_type=ContentType.xml_content.value,
                compress=True,
                content_encoding=ContentEncoding.gzip.value,
            )
        except Exception as e:
            logger.exception(f"Failed to save xml text to s3 due to {e}")
            return None

    def save_xml_to_s3(self, pdf_content: bytes, s3_location: S3Location):
        """
        pdf_content tends to come from:
            PyMuPdf.Document().write()
        """
        try:
            return self.save_document_content(
                file_contents=pdf_content,
                s3_location=s3_location,
                content_type=ContentType.pdf_content.value,
                compress=True,
                content_encoding=ContentEncoding.gzip.value,
            )
        except Exception as e:
            logger.exception(f"Failed to save pdf to s3 due to {e}")
            return None

    def save_dict_to_s3(self, content: Dict[str, Any], s3_location: S3Location):
        """ """
        try:
            file_contents = bytes(json.dumps(content).encode("UTF-8"))
            return self.save_document_content(
                file_contents=file_contents,
                s3_location=s3_location,
                content_type=ContentType.json_content.value,
                compress=True,
                content_encoding=ContentEncoding.gzip.value,
            )
        except Exception as e:
            logger.exception(f"Failed to save dict to s3 due to {e}")
            return None



from __future__ import annotations
import boto3
from botocore.config import Config
import json
import time
import os
import io
from typing import Optional, List, Dict
from enum import Enum
import pymupdf
import concurrent.futures
from dataclasses import dataclass

class PromptType(str, Enum):
    transaction_headers = "transactions_headers_prompt_v2.txt"
    transactions = "transactions_prompt.txt"
    # json = "json_system_prompt.txt"
    markdown = "markdown_system_prompt.txt"


@dataclass
class TokenUsage:
    input_tokens: int
    output_tokens: int
    total_tokens: int

    @classmethod
    def from_dict(cls, data: Dict[str, int]) -> TokenUsage:
        return cls(
            input_tokens=data.get('inputTokens', 0),
            output_tokens=data.get('outputTokens', 0),
            total_tokens=data.get('totalTokens', 0),
        )


@dataclass
class OCRResult:
    content: List[Dict[str, str]]
    token_usage: TokenUsage
    page_number: int

    @classmethod
    def from_dict(cls, data: Dict) -> OCRResult:
        return cls(
            content = data.get("content", []),
            token_usage = TokenUsage.from_dict(data.get("token_usage", {})),
            page_number = data.get("page_number", 0)
        )


class Bedrock:
    def __init__(
        self,
        model_id: str = "anthropic.claude-3-5-sonnet-20241022-v2:0",
        logger = None,
    ):

        self.session = Bedrock._set_session_params()
        self.logger = logger
        region_name = "ap-southeast-2"
        if self.session is None:
            self.session = boto3.Session(region_name = region_name)

        
        custom_config = Config(
            retries={
                'max_attempts': 2,  # Total attempts = 1 initial + 1 retry
                'mode': 'standard',  # or 'adaptive'
            }
        )      
        self.client = self.session.client("bedrock-runtime", region_name=region_name, config = custom_config)
        self.model_id = model_id

    @staticmethod
    def _set_session_params():
        try:
            aws_access_key_id = os.environ["AWS_ACCESS_KEY_ID"]
            aws_secret_access_key = os.environ["AWS_SECRET_ACCESS_KEY"]
            aws_session_token = os.environ["AWS_SESSION_TOKEN"]
            region_name = os.environ["AWS_DEFAULT_REGION"]
            return boto3.Session(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                region_name=region_name
            )
        except Exception as e:
            print(e)
            return None

    def _get_prompt(self, prompt_type: str) -> Optional[str]:
        if prompt_type not in list(PromptType):
            raise Exception(f"Error: Invalid prompt type")
        
        path = os.path.join(os.path.dirname(__file__), "prompts", prompt_type)
        try:
            with open(path, "r") as f:
                prompt = f.read()
                return prompt
        except Exception as e:
            self.logger.exception(f"Failed to get {prompt_type} prompt due to {e}")
        return None

    def _ocr(
        self, 
        prompt: str, 
        image_bytes: bytes,
        page_number: Optional[int] = 0
    ) -> Optional[OCRResult]:
        system_prompt = [{"text": prompt}]
        message = [
            {
                "role": "user",
                "content": [
                    {
                        "image": {
                            "format": "png",
                            "source": {
                                "bytes": image_bytes,
                            }
                        }
                    }
                ]
            }
        ]
        retries = 3
        for i in range(retries):
            self.logger.info(f"Attempt number {i} for {self.model_id} converse")
            try:
                response = self.client.converse(modelId = self.model_id, messages = message, system = system_prompt)
                if response['ResponseMetadata']['HTTPStatusCode'] == 200:
                    break
            except Exception as e:
                self.logger.exception(f"Error during conversation due to {e}")
                if i >= len(retries) - 1: raise Exception(e)
                continue
        
        result = {}
        result["content"] = json.loads(response["output"]["message"]["content"][0]["text"])
        result["token_usage"] = response["usage"]
        result["page_number"] = page_number
        return OCRResult.from_dict(data = result)
    
    def _parallel_ocr(
            self, 
            image_bytes_list: List[bytes], 
            prompt_type: str,
            max_workers: int = 10,
    ):
        results = list()
        prompt = self._get_prompt(prompt_type=prompt_type)
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            self.logger.info("Some log")
            execution_futures = {
                executor.submit(
                    self._ocr, 
                    prompt = prompt,
                    image_bytes = img,
                ): img for img in image_bytes_list
            }
            for future in concurrent.futures.as_completed(execution_futures):
                result = future.result()
                if result:
                    results.append(result)
        return results
    
    def get_ocr_result(
        self,
        pdf_bytes: io.BytesIO, 
        prompt_type: str,
        zoom: int = 7,
    ) -> List[OCRResult]:
        try:
            self.logger.info("Getting OCR Results")
            document = pymupdf.open(stream=pdf_bytes, filetype="pdf")
            pages: List[pymupdf.Page] = [p for p in document]

            image_bytes_list: List[bytes] = list()
            for i, p in enumerate(pages):
                try:
                    image_bytes: bytes = p.get_pixmap(matrix = pymupdf.Matrix(zoom, zoom)).tobytes("png")
                    image_bytes_list.append(image_bytes)
                except Exception as e:
                    self.logger.error(f"Could not get pix map for page {i}")
                    continue
            header_ocr_result = None
            if len(image_bytes_list) > 1:
                headers_prompt = self._get_prompt(prompt_type=PromptType.transaction_headers.value)
                for i in range(2):
                    # try to get headers from the first or second page
                    header_ocr_result = self._ocr(prompt=headers_prompt, image_bytes=image_bytes_list[i])
                    if header_ocr_result is None:
                        self.logger.info(f"No ocr result returned when getting headers {PromptType.transaction_headers.value}")
                    headers = header_ocr_result.content.get("headers")
                    if (len(headers) < 1) or (headers is None):
                        continue
                    else:
                        break

            transactions_prompt = self._get_prompt(prompt_type=prompt_type)
            if header_ocr_result:
                transactions_prompt = transactions_prompt.replace("#### TABLE HEADERS ####", json.dumps(header_ocr_result.content))

            self.logger.info("Got Prompt")
            results = list()

            for i, image_bytes in enumerate(image_bytes_list):
                self.logger.info(f"Starting OCR for page: {i}")
                results.append(self._ocr(image_bytes=image_bytes, prompt=transactions_prompt))
            return results
        except Exception as e:
            self.logger.exception(e)
            return []
            


import logging
import os


def select_powertools_logger(service_name: str) -> logging.Logger:
    """
    Returns the powertools logger if it can be found,
    Returns a Logger with name = service_name if powertools logger is not found
    """
    existing_loggers = [name for name in logging.root.manager.loggerDict]
    powertools_service_name = os.environ.get("POWERTOOLS_SERVICE_NAME")
    if powertools_service_name is not None:
        logger = (
            logging.getLogger(powertools_service_name)
            if powertools_service_name in existing_loggers
            else None
        )
        if logger:
            return logger
    return logging.getLogger(service_name)



from __future__ import annotations
from typing import Optional
from datetime import datetime
from enum import Enum
from uuid import uuid4
from copy import copy
from boto3.dynamodb.conditions import Key
from my_aws_helpers.dynamo import BaseTableObject, DynamoSerialiser, BaseQueries, Dynamo

class EventDynamoKeys:
    @staticmethod
    def get_event_pk(id: str):
        return f"id##{id}"
    
    @staticmethod
    def get_event_sk(id: str):
        return f"id##{id}"    


class EventStatus(str, Enum):
    in_progress = "in_progress"
    success = "success"
    fail = "fail"


class Event(BaseTableObject):
    status: str
    message: str
    id: str
    created_timestamp: datetime

    def __init__(
        self,
        status: str,
        message: str,
        id: str = None,
        created_timestamp: datetime = None,
    ):
        super().__init__()
        self.status: str = self.set_status(status=status)
        self.message: str = message
        self.id: str = id if id else uuid4().hex
        self.created_timestamp: datetime = created_timestamp if created_timestamp else datetime.now()

    def set_status(self, status: str) -> str:
        if status not in list(EventStatus):
            raise Exception("Status must be a member of EventStatus")
        return status

    def _get_pk(self):
        return EventDynamoKeys.get_event_pk(self.id)
    
    def _get_sk(self):
        return EventDynamoKeys.get_event_sk(self.id)  

    @classmethod
    def _from_dynamo_representation(cls, obj: dict):
        """
        Deserialises this object from Dynamo Representation
        """
        this_obj = copy(obj)
        this_obj.pop('pk')
        this_obj.pop('sk')
        return cls(**this_obj)

    def _to_dynamo_representation(self):
        """
        Serialises this object to Dynamo Representation
        """
        obj = copy(vars(self))
        obj['pk'] = self._get_pk()
        obj['sk'] = self._get_sk()
        return DynamoSerialiser.object_serialiser(obj=obj)


class EventQueries(BaseQueries):
    def __init__(self, table_name, client: Optional[Dynamo] = None):
        super().__init__(table_name, client)

    def put_event(self, event: Event) -> Event:
        response = self.client.put_item(item = event._to_dynamo_representation())
        if response["ResponseMetadata"]["HTTPStatusCode"] == 200:
            return event
        raise Exception(f"Failed to put object into dynamo table")
    
    def get_event_by_id(self, id: str):
        pk = EventDynamoKeys.get_event_pk(id=id)
        sk = EventDynamoKeys.get_event_sk(id=id)
        results = self.client.table.query(
            KeyConditionExpression=Key("pk").eq(pk) & Key("sk").eq(sk)
        )
        if len(results["Items"]) < 1:
            return None
        if len(results["Items"]) > 1:
            raise Exception(f"Should only be 1 event with id {id}")
        return Event._from_dynamo_representation(obj = results["Items"][0])

